/*
 * Win32 DirectX Wrappers, to simplify (?) my life.
 *
 * Copyright 1999 Brian King, under GNU Public License
 *
 */
#ifndef _MSC_VER
#define INITGUID
#endif

#include "config.h"
#include "sysconfig.h"

#include <stdlib.h>
#include <stdarg.h>
#include <signal.h>
#include <windows.h>

#ifdef _MSC_VER
#include <mmsystem.h>
#include <ddraw.h>
#include <dsound.h>
#else
#include "winstuff.h"
#endif

#include "sysdeps.h"
#include "options.h"
#include "picasso96.h"
#include "osdep/dxwrap.h"
#include "osdep/win32.h"
#include "osdep/win32gfx.h"

#define NOVSYNC

extern HRESULT (WINAPI * pDirectDrawCreate) (GUID FAR *, LPDIRECTDRAW FAR *, IUnknown FAR *);

#ifdef __GNUC__
/* We'll only QueryInterface for the highest versions */
DEFINE_GUID( IID_IDirectDraw4,        0x9c59509a,0x39bd,0x11d1,0x8c,0x4a,0x00,0xc0,0x4f,0xd9,0x30,0xc5 );
DEFINE_GUID( IID_IDirectDrawSurface4, 0x0B2B8630,0xAD35,0x11D0,0x8E,0xA6,0x00,0x60,0x97,0x97,0xEA,0x5B );
#endif

static BOOL bColourKeyAvailable = FALSE;
static BOOL bOverlayAvailable   = FALSE;
static DDCAPS drivercaps, helcaps;
static DWORD overlayflags;
static DDOVERLAYFX overlayfx;
extern COLORREF g_dwBackgroundColor;

/*
 * FUNCTION:   ShowDDCaps
 *
 * PURPOSE:    print out the DirectDraw Capabilities
 *
 * PARAMETERS:
 *   caps    - DDCAPS structure
 *   hw      - flag indicating if this 'caps' is for real hardware or the HEL
 *
 * RETURNS:    none
 *
 * NOTES:      none
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
static void ShowDDCaps( DDCAPS caps, int hw )
{
    static int shown = 0;
    static BOOL reset_shown = FALSE;
    if( currprefs.win32_logfile && shown >= 4 && !reset_shown)
    {
	shown = 0;
	reset_shown = TRUE;
    }

    if( shown < 2)
    {
	write_log( "DirectDraw Capabilities for %s:\n", hw ? "Display Driver Hardware" : "Display Driver Emulation Layer" );
	if( caps.dwCaps & DDCAPS_BLT )
	    write_log( "DDCAPS_BLT - Capable of blitting\n" );
	if( caps.dwCaps & DDCAPS_BLTCOLORFILL )
	    write_log( "DDCAPS_BLTCOLORFILL - Color filling with blitter\n" );
	if( caps.dwCaps & DDCAPS_BLTSTRETCH )
	    write_log( "DDCAPS_BLTSTRETCH - Stretch blitting\n" );
	if( caps.dwCaps & DDCAPS_CANBLTSYSMEM )
	    write_log( "DDCAPS_CANBLTSYSMEM - Blits from system memory\n" );
	if( caps.dwCaps & DDCAPS_CANCLIP )
	    write_log( "DDCAPS_CANCLIP - Can clip while blitting\n" );
	if( caps.dwCaps & DDCAPS_CANCLIPSTRETCHED )
	    write_log( "DDCAPS_CANCLIPSTRETCHED - Can clip while stretch-blitting\n" );
	if( caps.dwCaps & DDCAPS_COLORKEY )
	{
	    write_log( "DDCAPS_COLORKEY - Can color-key with blits/overlays\n" );
	    bColourKeyAvailable = TRUE;
	}
	if( caps.dwCaps & DDCAPS_GDI )
	    write_log( "DDCAPS_GDI - Display h/w shared with GDI\n" );
	if( caps.dwCaps & DDCAPS_NOHARDWARE )
	    write_log( "DDCAPS_NOHARDWARE - no h/w support!\n" );
	if( caps.dwCaps & DDCAPS_OVERLAY )
	{
	    write_log( "DDCAPS_OVERLAY - support for %d overlay(s)\n", caps.dwMaxVisibleOverlays );
	    if( bColourKeyAvailable )
	    {
		if( caps.dwCKeyCaps & DDCKEYCAPS_DESTOVERLAY )
		{
		    write_log( "DDCKEYCAPS_DESTOVERLAY - colour-keyed overlays\n" );
		    bOverlayAvailable = TRUE;
		}
	    }
	}
	if( caps.dwCaps & DDCAPS_OVERLAYFOURCC )
	    write_log( "DDCAPS_OVERLAYFOURCC - overlay can do color-space conversions\n" );
	if( caps.dwCaps & DDCAPS_OVERLAYSTRETCH )
	    write_log( "DDCAPS_OVERLAYSTRETCH - overlay can stretch with min=%d/max=%d\n", caps.dwMinOverlayStretch, caps.dwMaxOverlayStretch );
	if( caps.dwCaps & DDCAPS_VBI )
	    write_log( "DDCAPS_VBI - h/w can generate a vertical-blanking interrupt\n" );
	if( caps.dwCaps2 & DDCAPS2_CERTIFIED )
	    write_log( "DDCAPS2_CERTIFIED - certified driver\n" );
	if( caps.dwCaps2 & DDCAPS2_CANRENDERWINDOWED )
	    write_log( "DDCAPS2_CANRENDERWINDOWED - GDI windows can be seen when in full-screen\n" );
	if( caps.dwCaps2 & DDCAPS2_NOPAGELOCKREQUIRED )
	    write_log( "DDCAPS2_NOPAGELOCKREQUIRED - no page locking needed for DMA blits\n" );
	if( caps.dwCaps2 & DDCAPS2_FLIPNOVSYNC )
	    write_log( "DDCAPS2_FLIPNOVSYNC - can pass DDFLIP_NOVSYNC to Flip calls\n" );
	
	write_log( "Video memory: %d/%d\n", caps.dwVidMemFree, caps.dwVidMemTotal );
    }
    shown++;
}

/*
 * FUNCTION:   DirectDraw_ErrorString
 *
 * PURPOSE:    Convert a DirectX error value to a text-string
 *
 * PARAMETERS:
 *   ddrval  - HRESULT of a failed call
 *
 * RETURNS:    ASCII string
 *
 * NOTES:      none
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
char *DirectDraw_ErrorString( HRESULT ddrval )
{
    switch (ddrval) 
    {
    case DDERR_ALREADYINITIALIZED:
	    return "This object is already initialized.";
    case DDERR_CANNOTATTACHSURFACE:
	    return "Cannot attach surface.";
    case DDERR_CANNOTDETACHSURFACE:
	    return "Cannot detach surface.";
    case DDERR_CURRENTLYNOTAVAIL:
	    return "Support unavailable.";
    case DDERR_EXCEPTION:
	    return "Unexpected exception.";
    case DDERR_GENERIC:
	    return "Undefined";
    case DDERR_HEIGHTALIGN:
	    return "Height needs to be aligned.";
    case DDERR_INCOMPATIBLEPRIMARY:
	    return "New params doesn't match existing primary surface.";
    case DDERR_INVALIDCAPS:
	    return "Device doesn't have requested capabilities.";
    case DDERR_INVALIDCLIPLIST:
	    return "Provided clip-list not supported.";
    case DDERR_INVALIDMODE:
	    return "Mode not supported.";
    case DDERR_INVALIDOBJECT:
	    return "Invalid object.";
    case DDERR_INVALIDPARAMS:
	    return "Invalid params.";
    case DDERR_INVALIDPIXELFORMAT:
	    return "Device doesn't support requested pixel format.";
    case DDERR_INVALIDRECT:
	    return "Invalid RECT.";
    case DDERR_LOCKEDSURFACES:
	    return "Surface locked.";
    case DDERR_NO3D:
	    return "No 3d capabilities.";
    case DDERR_NOALPHAHW:
	    return "No alpha h/w.";
    case DDERR_NOCLIPLIST:
	    return "No clip-list.";
    case DDERR_NOCOLORCONVHW:
	    return "No colour-conversion h/w.";
    case DDERR_NOCOOPERATIVELEVELSET:
        return "No cooperative-level set.";
    case DDERR_NOCOLORKEY:
	    return "No colour-key.";
    case DDERR_NOCOLORKEYHW:
        return "No colour-key hardware.";
    case DDERR_NODIRECTDRAWSUPPORT:
        return "No DirectDraw support with this display driver!";
    case DDERR_NOEXCLUSIVEMODE:
        return "Exlusive-mode needed but not set yet.";
    case DDERR_NOFLIPHW:
        return "No flipping hardware.";
    case DDERR_NOGDI:
        return "No GDI present.";
    case DDERR_NOMIRRORHW:
        return "No mirror hardware.";
    case DDERR_NOTFOUND:
        return "Requested item not found.";
    case DDERR_NOOVERLAYHW:
        return "No overlay hardware.";
    case DDERR_NORASTEROPHW:
        return "No raster-op hardware.";
    case DDERR_NOROTATIONHW:
        return "No rotation hardware.";
    case DDERR_NOSTRETCHHW:
        return "No stretch hardware.";
    case DDERR_NOT4BITCOLOR:
        return "Not a 4-bit colour palette.";
    case DDERR_NOT4BITCOLORINDEX:
        return "Not a 4-bit colour-index.";
    case DDERR_NOT8BITCOLOR:
        return "Not an 8-bit colour palette.";
    case DDERR_NOTEXTUREHW:
        return "No texture hardware.";
    case DDERR_NOVSYNCHW:
        return "No vsync hardware.";
    case DDERR_NOZBUFFERHW:
        return "No z-buffer hardware.";
    case DDERR_NOZOVERLAYHW:
        return "No z-overlay hardware.";
    case DDERR_OUTOFCAPS:
    	return "out of caps";
    case DDERR_OUTOFMEMORY:
	    return "Out of memory.";
    case DDERR_OUTOFVIDEOMEMORY:
	    return "out of video memory.";
    case DDERR_OVERLAYCANTCLIP:
        return "no clipped overlay support in h/w.";
    case DDERR_OVERLAYCOLORKEYONLYONEACTIVE:
        return "can only have one color-key active at one time for overlays.";
    case DDERR_PALETTEBUSY:
    	return "Palette busy.";
    case DDERR_COLORKEYNOTSET:
    	return "No source colour-key provided.";
    case DDERR_SURFACEALREADYATTACHED:
        return "Surface already attached.";
    case DDERR_SURFACEALREADYDEPENDENT:
        return"Surface already dependent.";
    case DDERR_SURFACEBUSY:
    	return "Surface busy.";
    case DDERR_CANTLOCKSURFACE:
    	return "Access to surface refused (no DCI Provider).";
    case DDERR_SURFACEISOBSCURED:
        return "Surface is obscured.";
    case DDERR_SURFACELOST:
    	return "Surface lost.";
    case DDERR_SURFACENOTATTACHED:
        return "Surface not attached.";
    case DDERR_TOOBIGHEIGHT:
        return "Height requested is too large.";
    case DDERR_TOOBIGSIZE:
        return "Size requested is too large.";
    case DDERR_TOOBIGWIDTH:
        return "Width requested is too large.";
    case DDERR_UNSUPPORTED:
        return "Unsupported.";
    case DDERR_UNSUPPORTEDFORMAT:
    	return "Unsupported format.";
    case DDERR_UNSUPPORTEDMASK:
        return "Unsupported bitmask in pixel-format request.";
    case DDERR_VERTICALBLANKINPROGRESS:
        return "Vertical blank in progress.";
    case DDERR_WASSTILLDRAWING:
    	return "Was still drawing.";
    case DDERR_XALIGN:
        return "rectangle needs to be x-aligned.";
    case DDERR_INVALIDDIRECTDRAWGUID:
        return "invalid GUID for direct-draw.";
    case DDERR_DIRECTDRAWALREADYCREATED:
        return "DirectDraw object already created for this process.";
    case DDERR_NODIRECTDRAWHW:
        return "driver does not support h/w DirectDraw.";
    case DDERR_PRIMARYSURFACEALREADYEXISTS:
        return "Primary surface already exists for this process.";
    case DDERR_NOEMULATION:
        return "DirectDraw s/w emulation not available.";
    case DDERR_REGIONTOOSMALL:
        return "GetClipList() region is too small.";
    case DDERR_CLIPPERISUSINGHWND:
        return "Trying to set a clip-list for a clipper object which is already attached to an HWND.";
    case DDERR_NOCLIPPERATTACHED:
        return "No clipper object attached to surface object.";
    case DDERR_NOHWND:
        return "No HWND provided for clipper notification, or for CooperativeLevel.";
    case DDERR_HWNDSUBCLASSED:
        return "CooperativeLevel HWND has been subclassed, preventing state restoration.";
    case DDERR_HWNDALREADYSET:
        return "CooperativeLevel HWND already set - can't be reset while surfaces or palettes exist.";
    case DDERR_NOPALETTEATTACHED:
        return "No palette object attached to this surface.";
    case DDERR_NOPALETTEHW:
        return "No h/w support for 16 or 256 colour palettes.";
    case DDERR_BLTFASTCANTCLIP:
    	return "Cannot use BLTFAST with Clipper attached to surface.";
    case DDERR_NOBLTHW:
        return "No blitter.";
    case DDERR_NODDROPSHW:
        return "No DirectDraw ROP h/w.";
    case DDERR_OVERLAYNOTVISIBLE:
        return "GetOverlayPosition called on a hidden overlay.";
    case DDERR_NOOVERLAYDEST:
        return "GetOverlayPosition called on an overlay that UpdateOverlay has never been called on.";
    case DDERR_INVALIDPOSITION:
        return "Position of overlay on destination is no longer legal.";
    case DDERR_NOTAOVERLAYSURFACE:
        return "Overlay member is called for a non-overlay surface.";
    case DDERR_EXCLUSIVEMODEALREADYSET:
        return "Attempt to set CooperativeLevel when already set to exclusive.";
    case DDERR_NOTFLIPPABLE:
        return "Surface not flippable.";
    case DDERR_CANTDUPLICATE:
        return "Can't duplicate primary or 3d surfaces, or surfaces that are implicitly created.";
    case DDERR_NOTLOCKED:
        return "Surface not locked.";
    case DDERR_CANTCREATEDC:
        return "Cannot create any more DCs.";
    case DDERR_NODC:
        return "No DC ever created for this surface.";
    case DDERR_WRONGMODE:
        return "Surface cannot be restored, because it was created in a different mode.";
    case DDERR_IMPLICITLYCREATED:
        return "Surface cannot be restored, because it is an implicitly created surface.";
    case DDERR_NOTPALETTIZED:
        return "Surface being used is not a palette-based surface.";
    case DDERR_UNSUPPORTEDMODE:
        return "The display is currently in an unsupported mode.";
    case DDERR_NOMIPMAPHW:
        return "Operation could not be carried out because there is no mip-map texture mapping hardware present or available.";
    case DDERR_INVALIDSURFACETYPE:
        return "The requested action could not be performed because the surface was of the wrong type.";
    case DDERR_NOOPTIMIZEHW:
        return "Device does not support optimized surfaces, therefore no video memory optimized surfaces.";
    case DDERR_NOTLOADED:
        return "Surface is an optimized surface, but has not yet been allocated any memory.";
    case DDERR_NOFOCUSWINDOW:
        return "Attempt was made to create or set a device window without first setting the focus window.";
    case DDERR_DCALREADYCREATED:
        return "A DC has already been returned for this surface. Only one DC can be retrieved per surface.";
    case DDERR_NONONLOCALVIDMEM:
        return "An attempt was made to allocate non-local video memory from a device that does not support non-local video memory.";
    case DDERR_CANTPAGELOCK:
    	return "Attempt to page-lock a surface failed.";
    case DDERR_CANTPAGEUNLOCK:
	    return "Attempt to page-unlock a surface failed.";
    case DDERR_NOTPAGELOCKED:
        return "An attempt was made to page unlock a surface with no outstanding page locks.";
    case DDERR_MOREDATA:
        return "There is more data available than the specified buffer size could hold.";
    case DDERR_VIDEONOTACTIVE:
        return "The video port is not active.";
    case DDERR_DEVICEDOESNTOWNSURFACE:
        return "Surfaces created by one direct draw device cannot be used directly by another direct draw device.";
    case DDERR_NOTINITIALIZED:
        return "An attempt was made to invoke an interface member of a DirectDraw object created by CoCreateInstance() before it was initialized.";
    }
    return "";
}

static struct DirectDrawSurfaceMapper DirectDrawState;

static int lockcnt = 0;

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
static int LockStub( surface_type_e type )
{
    int result = 0;
    HRESULT ddrval;
    LPDIRECTDRAWSURFACE4 surface;
    LPDDSURFACEDESC2 surfacedesc;

    switch( type )
    {
    case primary_surface:
	surface = DirectDrawState.primary.surface_dx6;
	surfacedesc = &DirectDrawState.primary.desc_dx6;
	break;
    case secondary_surface:
	surface = DirectDrawState.secondary.surface_dx6;
        surfacedesc = &DirectDrawState.secondary.desc_dx6;
	break;
    case overlay_surface:
	surface = DirectDrawState.overlay.surface_dx6;
        surfacedesc = &DirectDrawState.overlay.desc_dx6;
	break;
    }

    if( lockcnt )
    {
#ifdef _DEBUG
	DebugBreak();
#endif
	return 1;
    }

    if( type == secondary_surface && DirectDrawState.flipping )
    {
	IDirectDrawSurface4_Restore( DirectDrawState.primary.surface_dx6 );
    }

    while ( (ddrval = IDirectDrawSurface4_Lock( surface, NULL, surfacedesc, DDLOCK_SURFACEMEMORYPTR | DDLOCK_WAIT,
                                                NULL ) ) != DD_OK )
    {
        if (ddrval == DDERR_SURFACELOST) {
    	    ddrval = IDirectDrawSurface4_Restore( surface );
            if (ddrval != DD_OK)
            {
		write_log( "Restore() failed - %s (%d)\n", DirectDraw_ErrorString (ddrval), (unsigned short) ddrval);
                result = 0;
                break;
            }
        }
        else if (ddrval != DDERR_SURFACEBUSY) 
        {
	    write_log ("lpDDS->Lock() failed - %s (%d)\n", DirectDraw_ErrorString (ddrval), (unsigned short) ddrval);
            result = 0;
            break;
        }
    }
    if( ddrval == DD_OK )
        result = 1;
    
    if( result )
        lockcnt++;

    return result;
}

/* For a given surface-type, update our DirectDrawState structure */
/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
int DirectDraw_SurfaceLock( surface_type_e surface_type )
{
    int result = 0;

    if( surface_type == lockable_surface )
        surface_type = DirectDraw_GetLockableType();

    switch( surface_type )
    {
        case primary_surface:
            DirectDrawState.primary.desc_dx6.dwSize = sizeof( DDSURFACEDESC2 );
            result = LockStub( surface_type );
        break;
        case secondary_surface:
            DirectDrawState.secondary.desc_dx6.dwSize = sizeof( DDSURFACEDESC2 );
            result = LockStub( surface_type );
        break;
	case overlay_surface:
	    DirectDrawState.overlay.desc_dx6.dwSize = sizeof( DDSURFACEDESC2 );
	    result = LockStub( surface_type );
        case lockable_surface:
        case invalid_surface:
        default:

        break;
    }
    DirectDrawState.locked = result;

    return result;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
char *DirectDraw_GetSurfacePointer( void )
{
    char *pixels = NULL;

    /* Make sure that somebody has done a lock before returning the lpSurface member */
    if( lockcnt )
    {
        pixels = DirectDrawState.lockable.lpdesc_dx6->lpSurface;
    }
    return pixels;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
LONG DirectDraw_GetSurfacePitch( void )
{
    LONG pitch = 0;

    pitch = DirectDrawState.lockable.lpdesc_dx6->lPitch;
    return pitch;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
DWORD DirectDraw_GetPixelFormatFlags( void )
{
    DWORD flags = 0;
    flags = DirectDrawState.lockable.lpdesc_dx6->ddpfPixelFormat.dwFlags;
    return flags;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
DWORD DirectDraw_GetSurfaceFlags( void )
{
    DWORD flags = 0;
    flags = DirectDrawState.lockable.lpdesc_dx6->dwFlags;
    return flags;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
DWORD DirectDraw_GetSurfaceBitCount( void )
{
    DWORD bits = 0;
    //?????JGI begin:
    if( DirectDrawState.lockable.lpdesc_dx6 )
    bits = DirectDrawState.lockable.lpdesc_dx6->ddpfPixelFormat.dwRGBBitCount;
    else
    bits = DirectDrawState.current.desc_dx6.ddpfPixelFormat.dwRGBBitCount;
    //?????JGI end.
    return bits;
}

/*
 * FUNCTION:   DirectDraw_GetPrimaryBitCount
 *
 * PURPOSE:    Return the bit-depth of the primary surface
 *
 * PARAMETERS: none
 *
 * RETURNS:    bit-depth
 *
 * NOTES:      
 *
 * HISTORY:
 *   2001.08.25  Brian King             Creation
 *
 */
DWORD DirectDraw_GetPrimaryBitCount( void )
{
    DWORD bits = 0;
    memset(&DirectDrawState.primary.desc_dx6,0,sizeof(DirectDrawState.primary.desc_dx6));
    DirectDrawState.primary.desc_dx6.dwSize = sizeof(DirectDrawState.primary.desc_dx6);

    IDirectDrawSurface2_GetSurfaceDesc(DirectDrawState.primary.surface_dx6, &DirectDrawState.primary.desc_dx6);
    bits = DirectDrawState.primary.desc_dx6.ddpfPixelFormat.dwRGBBitCount;
    return bits;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
DWORD DirectDraw_GetPixelFormatBitMask( DirectDraw_Mask_e mask )
{
    DWORD result = 0;
    switch( mask )
    {
        case red_mask:
            result = DirectDrawState.lockable.lpdesc_dx6->ddpfPixelFormat.dwRBitMask;
        break;
        case green_mask:
            result = DirectDrawState.lockable.lpdesc_dx6->ddpfPixelFormat.dwGBitMask;
        break;
        case blue_mask:
            result = DirectDrawState.lockable.lpdesc_dx6->ddpfPixelFormat.dwBBitMask;
        break;
    }
    return result;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
surface_type_e DirectDraw_GetLockableType( void )
{
    return DirectDrawState.surface_type;
}

/*
 * FUNCTION:   DirectDraw_IsLocked
 *
 * PURPOSE:    Return whether we're currently locked or unlocked
 *
 * PARAMETERS: none
 *
 * RETURNS:    TRUE if already locked, FALSE otherwise
 *
 * NOTES:      Used by DX_Blit to possibly unlock during Blit operation
 *
 * HISTORY:
 *   2000.04.30  Brian King             Creation
 *
 */
BOOL DirectDraw_IsLocked( void )
{
    return DirectDrawState.locked ? TRUE : FALSE;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
static surface_type_e try_surface_locks( int want_fullscreen )
{
    surface_type_e result = invalid_surface;

    if( DirectDrawState.isoverlay && DirectDraw_SurfaceLock( overlay_surface ) )
    {
	result = overlay_surface;
	write_log( "try_surface_locks() returning overlay\n" );
    }
    else if( want_fullscreen && WIN32GFX_IsPicassoScreen() )
    {
	if( DirectDraw_SurfaceLock( primary_surface ) )
	{
	    result = primary_surface;
	    write_log( "try_surface_locks() returning primary\n" );
	}
	else if( DirectDraw_SurfaceLock( secondary_surface ) )
	{
	    result = secondary_surface;
	    write_log( "try_surface_locks() returning secondary\n" );
	}
    }
    else
    {
        if( DirectDraw_SurfaceLock( secondary_surface ) )
        {
            result = secondary_surface;
	    write_log( "try_surface_locks() returning secondary\n" );
        }
    }

    return result;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      Named this way for historical reasons
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
void unlockscr( void )
{
    if( lockcnt > 0 )
    {
	lockcnt--;
	IDirectDrawSurface4_Unlock( DirectDrawState.lockable.surface_dx6,
				    DirectDrawState.lockable.lpdesc_dx6->lpSurface );
	DirectDrawState.locked = FALSE;
    }
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
int DirectDraw_Start( void )
{
    HRESULT ddrval;
    DDCAPS  drivercaps, helcaps;

    /* Prepare our DirectDrawState structure */
    ZeroMemory( &DirectDrawState, sizeof( DirectDrawState ) );

    ZeroMemory( &drivercaps, sizeof( drivercaps ) );
    ZeroMemory( &helcaps, sizeof( helcaps ) );
    drivercaps.dwSize = sizeof( DDCAPS );
    helcaps.dwSize = sizeof( DDCAPS );

    ddrval = DirectDrawCreate( NULL, &DirectDrawState.directdraw.dd, NULL );
    if (ddrval != DD_OK)
	goto oops;

    DirectDrawState.initialized = TRUE;

    ddrval = IDirectDraw_QueryInterface( DirectDrawState.directdraw.dd,
                                         &IID_IDirectDraw4,
                                         (LPVOID *)&DirectDrawState.directdraw.dd_dx6 );
    if( ddrval != DD_OK )
    {
        write_log( "System doesn't have DirectX-6 or higher... reverting to DirectX-3 style!\n" );
	ddrval = -1;
	/* @@@ TO DO @@@ */
    }

    DirectDraw_GetCaps( &drivercaps, &helcaps );
    ShowDDCaps( drivercaps, 1 );
    ShowDDCaps( helcaps, 0 );

    ddrval = DirectDraw_GetDisplayMode();
    if (ddrval != DD_OK)
	goto oops;

    return 1;

  oops:
    gui_message("start_ddraw(): DirectDraw initialization failed with %s/%d\n", DirectDraw_ErrorString (ddrval), ddrval);
    DirectDraw_Release();
    return 0;
}

#define releaser(x,y) if( x ) { y( x ); x = NULL; }

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
void DirectDraw_Release( void )
{
    releaser( DirectDrawState.lpDDC, IDirectDrawClipper_Release );
    releaser( DirectDrawState.lpDDP, IDirectDrawPalette_Release );
    releaser( DirectDrawState.primary.surface_dx6, IDirectDrawSurface4_Release );
    releaser( DirectDrawState.primary.surface, IDirectDrawSurface_Release );
    releaser( DirectDrawState.secondary.surface_dx6, IDirectDrawSurface4_Release );
    releaser( DirectDrawState.overlay.surface_dx6, IDirectDrawSurface4_Release );
    releaser( DirectDrawState.overlay.surface, IDirectDrawSurface_Release );

    if( !DirectDrawState.flipping )
	releaser( DirectDrawState.secondary.surface, IDirectDrawSurface_Release );

    if( DirectDrawState.directdraw.dd_dx6 )
    {
	IDirectDraw4_RestoreDisplayMode( DirectDrawState.directdraw.dd_dx6 );
	IDirectDraw4_SetCooperativeLevel( DirectDrawState.directdraw.dd_dx6, hAmigaWnd, DDSCL_NORMAL);
    }
    releaser( DirectDrawState.directdraw.dd_dx6, IDirectDraw4_Release );

    if( DirectDrawState.directdraw.dd )
    {
	IDirectDraw_RestoreDisplayMode( DirectDrawState.directdraw.dd );
	IDirectDraw_SetCooperativeLevel( DirectDrawState.directdraw.dd, hAmigaWnd, DDSCL_NORMAL);
    }
    releaser( DirectDrawState.directdraw.dd, IDirectDraw_Release );

    DirectDrawState.lockable.lpdesc = NULL;
    DirectDrawState.lockable.lpdesc_dx6 = NULL;
    DirectDrawState.lockable.surface = NULL;
    DirectDrawState.lockable.surface_dx6 = NULL;

    DirectDrawState.surface_type = invalid_surface;
    DirectDrawState.initialized = FALSE;
    DirectDrawState.isoverlay = FALSE;
}

/*
 * FUNCTION:   DirectDraw_SetCooperativeLevel
 *
 * PURPOSE:    Wrapper for setting the cooperative level (fullscreen or normal)
 *
 * PARAMETERS:
 *   window		Window to set the cooperative level for
 *   want_fullscreen	fullscreen mode flag
 *
 * RETURNS:		result of underlying DirectDraw call
 *
 * NOTES:		Updates the .fullscreen and .window members.
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_SetCooperativeLevel( HWND window, int want_fullscreen )
{
    HRESULT ddrval;

    ddrval = IDirectDraw4_SetCooperativeLevel( DirectDrawState.directdraw.dd_dx6,
                                               window,
                                               want_fullscreen ?
                                               DDSCL_ALLOWREBOOT | DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN :
                                               DDSCL_NORMAL );
    if( ddrval == DD_OK )
    {
	DirectDrawState.fullscreen = want_fullscreen;
	DirectDrawState.window = window;
    }
    return ddrval;
}

/*
 * FUNCTION:   DirectDraw_GetCooperativeLevel
 *
 * PURPOSE:    Wrapper for setting the cooperative level (fullscreen or normal)
 *
 * PARAMETERS:
 *   window		Window to set the cooperative level for
 *   want_fullscreen	fullscreen mode flag
 *
 * RETURNS:		result of underlying DirectDraw call
 *
 * NOTES:		Updates the .fullscreen and .window members.
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
BOOL DirectDraw_GetCooperativeLevel( HWND *window, int *fullscreen )
{
    BOOL result = FALSE;

    if( DirectDrawState.initialized )
    {
	*fullscreen = DirectDrawState.fullscreen;
	*window = DirectDrawState.window;
	result = TRUE;
    }
    return result;
}

/*
 * FUNCTION:   DirectDraw_SetDisplayMode
 *
 * PURPOSE:    Change the display-mode to width x height pixels, with a given
 *             vertical refresh-rate.
 *
 * PARAMETERS:
 *   width   - width of display in pixels
 *   height  - height of display in pixels
 *   freq    - vertical refresh-rate in Hz
 *
 * RETURNS:
 *   ddrval  - HRESULT indicating success (DD_OK) or failure
 *
 * NOTES:      The freq parameter is only obeyed on when we're using DirectX 6
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_SetDisplayMode( int width, int height, int bits, int freq )
{
    HRESULT ddrval;

    ddrval = IDirectDraw4_SetDisplayMode( DirectDrawState.directdraw.dd_dx6,
                                          width, height, bits, freq, 0 );
    return ddrval;
}

/*
 * FUNCTION:   DirectDraw_GetDisplayMode
 *
 * PURPOSE:    Get the display-mode characteristics.
 *
 * PARAMETERS: none
 *
 * RETURNS:
 *   ddrval  - HRESULT indicating success (DD_OK) or failure
 *
 * NOTES:      none
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_GetDisplayMode( void )
{
    HRESULT ddrval;

    DirectDrawState.current.desc_dx6.dwSize = sizeof( DDSURFACEDESC2 );
    ddrval = IDirectDraw4_GetDisplayMode( DirectDrawState.directdraw.dd_dx6,
                                          &DirectDrawState.current.desc_dx6 );

    /* We fill in the current.desc in all cases */
    DirectDrawState.current.desc.dwSize = sizeof( DDSURFACEDESC );
    ddrval = IDirectDraw_GetDisplayMode( DirectDrawState.directdraw.dd,
                                         &DirectDrawState.current.desc );
    return ddrval;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_GetCaps( DDCAPS *driver_caps, DDCAPS *hel_caps )
{
    HRESULT ddrval;

    ddrval = IDirectDraw4_GetCaps( DirectDrawState.directdraw.dd_dx6,
                                   driver_caps, hel_caps );
    return ddrval;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_CreateClipper( void )
{
    HRESULT ddrval;
    ddrval = IDirectDraw4_CreateClipper( DirectDrawState.directdraw.dd_dx6,
                                         0,
                                         &DirectDrawState.lpDDC,
                                         NULL );
    return ddrval;
}

static DWORD ConvertGDIColor( COLORREF dwGDIColor )
{
    COLORREF rgbT;
    HDC hdc;
    DWORD dw = CLR_INVALID;
    DDSURFACEDESC2 ddsd,pdds;
    HRESULT hr;

    memset(&pdds,0,sizeof(pdds));
    pdds.dwSize = sizeof(pdds);
 
    IDirectDrawSurface2_GetSurfaceDesc(DirectDrawState.primary.surface_dx6, &pdds);

    //  Use GDI SetPixel to color match for us
    if( dwGDIColor != CLR_INVALID && IDirectDrawSurface2_GetDC(DirectDrawState.primary.surface_dx6, &hdc) == DD_OK)
    {
        rgbT = GetPixel(hdc, 0, 0);     // Save current pixel value
        SetPixel(hdc, 0, 0, dwGDIColor);       // Set our value
        IDirectDrawSurface2_ReleaseDC(DirectDrawState.primary.surface_dx6,hdc);
    }

    // Now lock the surface so we can read back the converted color
    ddsd.dwSize = sizeof(ddsd);
    hr = IDirectDrawSurface2_Lock(DirectDrawState.primary.surface_dx6, NULL, &ddsd, DDLOCK_WAIT, NULL );
    if( hr == DD_OK)
    {
        dw = *(DWORD *) ddsd.lpSurface; 
        if( ddsd.ddpfPixelFormat.dwRGBBitCount < 32 ) // Mask it to bpp
            dw &= ( 1 << ddsd.ddpfPixelFormat.dwRGBBitCount ) - 1;
	IDirectDrawSurface2_Unlock(DirectDrawState.primary.surface_dx6,NULL);
    }

    //  Now put the color that was there back.
    if( dwGDIColor != CLR_INVALID && IDirectDrawSurface2_GetDC(DirectDrawState.primary.surface_dx6,&hdc) == DD_OK )
    {
        SetPixel( hdc, 0, 0, rgbT );
	IDirectDrawSurface2_ReleaseDC(DirectDrawState.primary.surface_dx6,hdc);
    }
    
    return dw;    
}


HRESULT DirectDraw_CreateOverlaySurface( int width, int height, int bits)
{
    DDSURFACEDESC ddsd;
    DDPIXELFORMAT ddpfOverlayFormat;
    HRESULT ddrval = DDERR_UNSUPPORTED;
    DWORD dwDDSColor;

    if( bOverlayAvailable )
    {
	write_log( "CreateOverlaySurface being called with %d-bits!\n", bits );
	if( bits == 16 )
	{
	    // Set the overlay format to 16 bit RGB 5:6:5
	    ZeroMemory( &ddpfOverlayFormat, sizeof(ddpfOverlayFormat) );
	    ddpfOverlayFormat.dwSize        = sizeof(ddpfOverlayFormat);
	    ddpfOverlayFormat.dwFlags       = DDPF_RGB;
	    ddpfOverlayFormat.dwRGBBitCount = 16;
	    ddpfOverlayFormat.dwRBitMask    = 0xF800; 
	    ddpfOverlayFormat.dwGBitMask    = 0x07E0;
	    ddpfOverlayFormat.dwBBitMask    = 0x001F; 
	}
	else if( bits == 32 )
	{
	    // Set the overlay format to 32 bit ARGB 8:8:8:8
	    ZeroMemory( &ddpfOverlayFormat, sizeof(ddpfOverlayFormat) );
	    ddpfOverlayFormat.dwSize        = sizeof(ddpfOverlayFormat);
	    ddpfOverlayFormat.dwFlags       = DDPF_RGB;
	    ddpfOverlayFormat.dwRGBBitCount = 32;
	    ddpfOverlayFormat.dwRBitMask    = 0x00FF0000; 
	    ddpfOverlayFormat.dwGBitMask    = 0x0000FF00;
	    ddpfOverlayFormat.dwBBitMask    = 0x000000FF; 
	}
	else if( bits == 8 )
	{
	    // Set the overlay format to 8 bit palette
	    ZeroMemory( &ddpfOverlayFormat, sizeof(ddpfOverlayFormat) );
	    ddpfOverlayFormat.dwSize        = sizeof(ddpfOverlayFormat);
	    ddpfOverlayFormat.dwFlags       = DDPF_RGB | DDPF_PALETTEINDEXED8;
	    ddpfOverlayFormat.dwRGBBitCount = 8;
	    ddpfOverlayFormat.dwRBitMask    = 0x00000000; 
	    ddpfOverlayFormat.dwGBitMask    = 0x00000000;
	    ddpfOverlayFormat.dwBBitMask    = 0x00000000; 
	}
	else
	{
	    // We don't handle this case...
	    return DDERR_INVALIDPIXELFORMAT;
	}
    
	// Setup the overlay surface's attributes in the surface descriptor
	ZeroMemory( &ddsd, sizeof(ddsd) );
	ddsd.dwSize            = sizeof(ddsd);
	ddsd.dwFlags           = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH | DDSD_PIXELFORMAT;
	ddsd.ddsCaps.dwCaps    = DDSCAPS_OVERLAY | DDSCAPS_VIDEOMEMORY;
	ddsd.dwWidth           = width;
	ddsd.dwHeight          = height;
	ddsd.ddpfPixelFormat   = ddpfOverlayFormat;

	ZeroMemory(&overlayfx,sizeof(overlayfx));
	overlayfx.dwSize = sizeof(overlayfx);
	overlayflags = DDOVER_SHOW | DDOVER_DDFX | DDOVER_KEYDESTOVERRIDE;

	dwDDSColor = ConvertGDIColor( g_dwBackgroundColor );
	overlayfx.dckDestColorkey.dwColorSpaceLowValue  = dwDDSColor;
	overlayfx.dckDestColorkey.dwColorSpaceHighValue = dwDDSColor;

	// Attempt to create the surface with theses settings
	ddrval = IDirectDraw_CreateSurface( DirectDrawState.directdraw.dd, &ddsd, &DirectDrawState.overlay.surface, NULL);
	if( ddrval == DD_OK )
	{
	    ddrval = IDirectDrawSurface_QueryInterface( DirectDrawState.overlay.surface,
			&IID_IDirectDrawSurface4,(LPVOID *)&DirectDrawState.overlay.surface_dx6 );
	    DirectDrawState.isoverlay = 1;
	}
	else
	{
	    DirectDrawState.isoverlay = 0;
	}
    }
    else
    {
	write_log( "CreateOverlaySurface being called, but no overlay support with this card...!\n" );
    }
    return ddrval;
}


/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_CreateSurface( int width, int height )
{
    HRESULT ddrval;

    DirectDrawState.flipping = FALSE;

    if( DirectDrawState.fullscreen ) // Create a flipping pair!
    {
	ZeroMemory( &DirectDrawState.primary.desc, sizeof( DDSURFACEDESC ) );
	DirectDrawState.primary.desc.dwSize = sizeof( DDSURFACEDESC );
	DirectDrawState.primary.desc.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
	DirectDrawState.primary.desc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_COMPLEX | DDSCAPS_FLIP;
	DirectDrawState.primary.desc.dwBackBufferCount = 1;
	ddrval = IDirectDraw_CreateSurface( DirectDrawState.directdraw.dd, 
					    &DirectDrawState.primary.desc,
					    &DirectDrawState.primary.surface,
					    NULL );
	if( ddrval != DD_OK )
	{
	    // Create a non-flipping pair, since the flipping pair creation failed...

	    ZeroMemory( &DirectDrawState.primary.desc, sizeof( DDSURFACEDESC ) );
	    DirectDrawState.primary.desc.dwSize = sizeof( DDSURFACEDESC );
	    DirectDrawState.primary.desc.dwFlags = DDSD_CAPS;
	    DirectDrawState.primary.desc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
	    ddrval = IDirectDraw_CreateSurface( DirectDrawState.directdraw.dd, 
						&DirectDrawState.primary.desc,
						&DirectDrawState.primary.surface,
						NULL );
	}
	else
	{
	    DirectDrawState.flipping = TRUE;
	}
    }
    else
    {
	// We're not full-screen, so you cannot create a flipping pair...

	ZeroMemory( &DirectDrawState.primary.desc, sizeof( DDSURFACEDESC ) );
	DirectDrawState.primary.desc.dwSize = sizeof( DDSURFACEDESC );
	DirectDrawState.primary.desc.dwFlags = DDSD_CAPS;
	DirectDrawState.primary.desc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
	ddrval = IDirectDraw_CreateSurface( DirectDrawState.directdraw.dd, 
					    &DirectDrawState.primary.desc,
					    &DirectDrawState.primary.surface,
					    NULL );
    }

    if( ddrval != DD_OK )
    {
	goto out;
    }
    else
    {
	write_log( "DDRAW: Primary %ssurface created in video-memory\n", DirectDrawState.flipping ? "flipping " : "" );
    }

    // Check if we can access the back-buffer of our flipping-pair (if present)
    if( DirectDrawState.flipping )
    {
	DDSCAPS ddSCaps;
	ZeroMemory(&ddSCaps, sizeof(ddSCaps));
	ddSCaps.dwCaps = DDSCAPS_BACKBUFFER;

	ddrval = IDirectDrawSurface_GetAttachedSurface( DirectDrawState.primary.surface, &ddSCaps, &DirectDrawState.secondary.surface );
	if( ddrval == DD_OK )
	{
	    // Get our IDirectDrawSurface4 pointer
	    ddrval = IDirectDrawSurface_QueryInterface( DirectDrawState.secondary.surface,
							&IID_IDirectDrawSurface4,
							(LPVOID *)&DirectDrawState.secondary.surface_dx6 );
	    if( ddrval != DD_OK )
	    {
		goto out;
	    }
	}
	else
	{
	    DirectDrawState.flipping = FALSE;
	}
    }

    // Get our IDirectDrawSurface4 pointer
    ddrval = IDirectDrawSurface_QueryInterface( DirectDrawState.primary.surface,
						&IID_IDirectDrawSurface4,
						(LPVOID *)&DirectDrawState.primary.surface_dx6 );
    if( ddrval != DD_OK )
    {
	goto out;
    }

    // We always want a secondary-buffer when creating our primary-surface.  If we're a flipping pair,
    // the secondary buffer is already allocated.  If we failed to create a flipping pair, or because
    // we're not full-screen, then lets create ourselves a back-buffer manually.
    if( !DirectDrawState.flipping )
    {
        ZeroMemory( &DirectDrawState.secondary.desc, sizeof( DDSURFACEDESC ) );
        DirectDrawState.secondary.desc.dwSize = sizeof( DDSURFACEDESC );
        DirectDrawState.secondary.desc.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;
        DirectDrawState.secondary.desc.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
        DirectDrawState.secondary.desc.dwWidth = width;
        DirectDrawState.secondary.desc.dwHeight = height;
        ddrval = IDirectDraw_CreateSurface( DirectDrawState.directdraw.dd, 
                                            &DirectDrawState.secondary.desc,
                                            &DirectDrawState.secondary.surface,
                                            NULL );
        if( ddrval != DD_OK )
        {
	    write_log( "DDRAW:Secondary surface creation attempt #1 failed with %s/%d\n", DirectDraw_ErrorString(ddrval), ddrval );
            DirectDrawState.secondary.desc.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | DDSCAPS_SYSTEMMEMORY;
            ddrval = IDirectDraw_CreateSurface( DirectDrawState.directdraw.dd, 
                                                &DirectDrawState.secondary.desc,
                                                &DirectDrawState.secondary.surface,
                                                NULL );
	    if( ddrval == DD_OK )
		write_log( "DDRAW: Secondary surface created in plain system-memory\n" );
	    else
	    {
		goto out;
	    }
        }
        else
        {
            write_log( "DDRAW: Secondary surface created in video-memory\n" );
        }

	// Get our IDirectDrawSurface4 pointer
        ddrval = IDirectDrawSurface_QueryInterface( DirectDrawState.secondary.surface,
                                                    &IID_IDirectDrawSurface4,
                                                    (LPVOID *)&DirectDrawState.secondary.surface_dx6 );
        if( ddrval != DD_OK )
        {
	    goto out;
        }
    }
out:
    return ddrval;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
void DirectDraw_ClearSurfaces( void )
{
    DDBLTFX ddbltfx;
    memset( &ddbltfx, 0, sizeof( ddbltfx ) );
    ddbltfx.dwFillColor = 0;
    ddbltfx.dwSize = sizeof( ddbltfx );

    DirectDraw_Blt( secondary_surface, NULL, invalid_surface, NULL, DDBLT_WAIT | DDBLT_COLORFILL, &ddbltfx );
    if( DirectDrawState.isoverlay )
	DirectDraw_Blt( overlay_surface, NULL, invalid_surface, NULL, DDBLT_WAIT | DDBLT_COLORFILL, &ddbltfx );
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
int DirectDraw_DetermineLocking( int wantfull )
{
    int result = 0;

    switch( DirectDrawState.surface_type = try_surface_locks( wantfull ) )
    {
    case invalid_surface:
    case lockable_surface:
        DirectDrawState.lockable.lpdesc = NULL;
        DirectDrawState.lockable.lpdesc_dx6 = NULL;
        DirectDrawState.lockable.surface = NULL;
        DirectDrawState.lockable.surface_dx6 = NULL;
        write_log( "set_ddraw: Couldn't lock primary, and no secondary available.\n" );
        break;
    case primary_surface:
        DirectDrawState.lockable.lpdesc = &DirectDrawState.primary.desc;
        DirectDrawState.lockable.lpdesc_dx6 = &DirectDrawState.primary.desc_dx6;
        DirectDrawState.lockable.surface = DirectDrawState.primary.surface;
        DirectDrawState.lockable.surface_dx6 = DirectDrawState.primary.surface_dx6;

	result = 1;
        break;
    case overlay_surface:
        DirectDrawState.lockable.lpdesc = &DirectDrawState.overlay.desc;
        DirectDrawState.lockable.lpdesc_dx6 = &DirectDrawState.overlay.desc_dx6;
        DirectDrawState.lockable.surface = DirectDrawState.overlay.surface;
        DirectDrawState.lockable.surface_dx6 = DirectDrawState.overlay.surface_dx6;
	result = 1;
        break;
    case secondary_surface:
        DirectDrawState.lockable.lpdesc = &DirectDrawState.secondary.desc;
        DirectDrawState.lockable.lpdesc_dx6 = &DirectDrawState.secondary.desc_dx6;
        DirectDrawState.lockable.surface = DirectDrawState.secondary.surface;
        DirectDrawState.lockable.surface_dx6 = DirectDrawState.secondary.surface_dx6;

	result = 1;
        break;
    }

    if( DirectDrawState.lockable.surface )
        DirectDraw_SurfaceUnlock();

    return result;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_SetClipper( HWND hWnd )
{
    HRESULT ddrval;

    ddrval = IDirectDrawSurface4_SetClipper( DirectDrawState.primary.surface_dx6,
                                             hWnd ? DirectDrawState.lpDDC : NULL );
    if( hWnd && ( ddrval == DD_OK ) )
    {
        ddrval = IDirectDrawClipper_SetHWnd( DirectDrawState.lpDDC, 0, hWnd );
    }
    return ddrval;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_GetClipList( LPRGNDATA cliplist, LPDWORD size )
{
    HRESULT ddrval;

    ddrval = IDirectDrawClipper_GetClipList( DirectDrawState.lpDDC, NULL, cliplist, size );

    return ddrval;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
BYTE DirectDraw_GetBytesPerPixel( void )
{
    int bpp;
    bpp = ( DirectDrawState.lockable.lpdesc_dx6->ddpfPixelFormat.dwRGBBitCount + 7 ) >> 3;
    return bpp;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_SetPalette( int remove )
{
    HRESULT ddrval;
    ddrval = IDirectDrawSurface4_SetPalette( DirectDrawState.primary.surface_dx6,
                                             remove ? NULL : DirectDrawState.lpDDP );
    return ddrval;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_CreatePalette( LPPALETTEENTRY pal )
{
    HRESULT ddrval;
    ddrval = IDirectDraw4_CreatePalette( DirectDrawState.directdraw.dd_dx6,
                                        DDPCAPS_8BIT | DDPCAPS_ALLOW256,
                                        pal,
                                        &DirectDrawState.lpDDP, NULL);
    if( ddrval == DD_OK )
    {
        ddrval = DirectDraw_SetPalette(0);
    }
    return ddrval;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_SetPaletteEntries( int start, int count, PALETTEENTRY *palette )
{
    HRESULT ddrval = DDERR_NOPALETTEATTACHED;
    if( DirectDrawState.lpDDP )
        ddrval = IDirectDrawPalette_SetEntries( DirectDrawState.lpDDP, 0, start, count, palette );
    return ddrval;
}

/* Return one of the pixel formats declared in picasso96.h if the surface
 * is usable for us, or RGBFB_NONE if it is not usable.  */
/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
RGBFTYPE DirectDraw_GetSurfacePixelFormat( LPVOID surface )
{
    int surface_is_dx6 = 0;
    DDPIXELFORMAT *pfp = NULL;
    DWORD r, g, b;
    DWORD surf_flags;
    
    LPDDSURFACEDESC lpsd = NULL;
    LPDDSURFACEDESC2 lpsd2 = NULL;

    /* Examine the size to see what type of surface-descriptor this is */
    if( ((LPDDSURFACEDESC)surface)->dwSize == sizeof( DDSURFACEDESC ) )
    {
        lpsd = (LPDDSURFACEDESC)surface;
        surf_flags = lpsd->dwFlags;
        pfp = &lpsd->ddpfPixelFormat;
    }
    else
    {
        lpsd2 = (LPDDSURFACEDESC2)surface;
        surf_flags = lpsd2->dwFlags;
        pfp = &lpsd2->ddpfPixelFormat;
    }

    if( ( surf_flags & DDSD_PIXELFORMAT ) == 0x0 )
	return RGBFB_NONE;

    if ((pfp->dwFlags & DDPF_RGB) == 0)
	return RGBFB_NONE;

    r = pfp->dwRBitMask;
    g = pfp->dwGBitMask;
    b = pfp->dwBBitMask;
    switch (pfp->dwRGBBitCount) {
     case 8:
	if ((pfp->dwFlags & DDPF_PALETTEINDEXED8) != 0)
	    return RGBFB_CHUNKY;
	break;

     case 16:
	if (r == 0xF800 && g == 0x07E0 && b == 0x001F)
	    return RGBFB_R5G6B5PC;
	if (r == 0x7C00 && g == 0x03E0 && b == 0x001F)
	    return RGBFB_R5G5B5PC;
	if (b == 0xF800 && g == 0x07E0 && r == 0x001F)
	    return RGBFB_B5G6R5PC;
	if (b == 0x7C00 && g == 0x03E0 && r == 0x001F)
	    return RGBFB_B5G5R5PC;
	/* This happens under NT - with r == b == g == 0 !!! */
	write_log ("Unknown 16 bit format %d %d %d\n", r, g, b);
	break;

     case 24:
	if (r == 0xFF0000 && g == 0x00FF00 && b == 0x0000FF)
	    return RGBFB_B8G8R8;
	if (r == 0x0000FF && g == 0x00FF00 && b == 0xFF0000)
	    return RGBFB_R8G8B8;
	break;

     case 32:
	if (r == 0x00FF0000 && g == 0x0000FF00 && b == 0x000000FF)
	    return RGBFB_B8G8R8A8;
	if (r == 0x000000FF && g == 0x0000FF00 && b == 0x00FF0000)
	    return RGBFB_R8G8B8A8;
	if (r == 0xFF000000 && g == 0x00FF0000 && b == 0x0000FF00)
	    return RGBFB_A8B8G8R8;
	if (r == 0x0000FF00 && g == 0x00FF0000 && b == 0xFF000000)
	    return RGBFB_A8R8G8B8;
	break;
	
     default:
	gui_message("Unknown %d bit format %d %d %d\n", pfp->dwRGBBitCount, r, g, b); /* %%% - BERND, and here too... */
	break;
    }
    return RGBFB_NONE;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
RGBFTYPE DirectDraw_GetPixelFormat( void )
{
    RGBFTYPE type;
    if( DirectDrawState.lockable.lpdesc_dx6 )
	type = DirectDraw_GetSurfacePixelFormat( DirectDrawState.lockable.lpdesc_dx6 );
    else
	type = DirectDraw_GetSurfacePixelFormat( &DirectDrawState.current.desc_dx6 );
    return type;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
DWORD DirectDraw_CurrentWidth( void )
{
    DWORD width;
    width = DirectDrawState.current.desc_dx6.dwWidth;
    return width;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
DWORD DirectDraw_CurrentHeight( void )
{
    DWORD height;
    height = DirectDrawState.current.desc_dx6.dwHeight;
    return height;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
static int DirectDraw_BltFastStub4( LPDIRECTDRAWSURFACE4 dstsurf, DWORD x, DWORD y, LPDIRECTDRAWSURFACE4 srcsurf, LPRECT srcrect )
{
    int result = 0;
    HRESULT ddrval;

    while( ( ddrval = IDirectDrawSurface4_BltFast( dstsurf, x, y, srcsurf, srcrect, DDBLTFAST_NOCOLORKEY | DDBLTFAST_WAIT ) ) != DD_OK )
    {
        if (ddrval == DDERR_SURFACELOST) 
        {
    	    ddrval = IDirectDrawSurface4_Restore( dstsurf );
            if (ddrval != DD_OK)
            {
                break;
            }
        }
        else if (ddrval != DDERR_SURFACEBUSY) 
        {
	    write_log("BltFastStub4(): DirectDrawSurface4_BltFast() failed with %s/%d\n", DirectDraw_ErrorString (ddrval), ddrval);
            break;
        }
    }
    if( ddrval == DD_OK )
    {
        result = 1;
    }
    return result;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
int DirectDraw_BltFast( surface_type_e dsttype, DWORD left, DWORD top, surface_type_e srctype, LPRECT srcrect )
{
    int result;

    LPDIRECTDRAWSURFACE4 lpDDS4_dst, lpDDS4_src;
    if( dsttype == primary_surface )
    {
        lpDDS4_dst = DirectDrawState.primary.surface_dx6;
    }
    else
    {
        lpDDS4_dst = DirectDrawState.secondary.surface_dx6;
    }
    if( srctype == primary_surface )
    {
        lpDDS4_src = DirectDrawState.primary.surface_dx6;
    }
    else
    {
        lpDDS4_src = DirectDrawState.secondary.surface_dx6;
    }
    result = DirectDraw_BltFastStub4( lpDDS4_dst, left, top, lpDDS4_src, srcrect );
    return result;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
static int DirectDraw_BltStub( LPDIRECTDRAWSURFACE4 dstsurf, LPRECT dstrect, LPDIRECTDRAWSURFACE4 srcsurf, LPRECT srcrect, DWORD flags, LPDDBLTFX ddbltfx )
{
    int result = 0;
    HRESULT ddrval;

    while( ( ddrval = IDirectDrawSurface4_Blt( dstsurf, dstrect, srcsurf, srcrect, flags, ddbltfx ) ) != DD_OK )
    {
        if (ddrval == DDERR_SURFACELOST) 
        {
    	    ddrval = IDirectDrawSurface4_Restore( dstsurf );
            if (ddrval != DD_OK)
            {
                break;
            }
        }
        else if (ddrval != DDERR_SURFACEBUSY) 
        {
	    write_log("BltStub(): DirectDrawSurface4_Blt() failed with %s/%d\n", DirectDraw_ErrorString (ddrval), ddrval);
            break;
        }
#if 0
	else
	{
	    write_log( "Blt() failed - %s (%d)\n", DirectDraw_ErrorString (ddrval), (unsigned short) ddrval);
	    result = 0;
	    break;
	}
#endif
    }
    if( ddrval == DD_OK )
    {
        result = 1;
    }
    return result;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
int DirectDraw_Flip( void )
{
    int result = 0;
    HRESULT ddrval;

    if( DirectDrawState.flipping )
    {
#ifdef NOVSYNC
	ddrval = IDirectDrawSurface4_Flip( DirectDrawState.primary.surface_dx6, NULL, DDFLIP_WAIT | DDFLIP_NOVSYNC);
#else
	ddrval = IDirectDrawSurface4_Flip( DirectDrawState.primary.surface_dx6, NULL, DDFLIP_WAIT );
#endif
	if( ddrval == DD_OK )
	{
	    result = 1;
	}
	else
	{
	    write_log("FLIP: DirectDrawSurface_Flip() failed with %s/%d\n", DirectDraw_ErrorString (ddrval), ddrval);
	}
    }

    return result;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
int DirectDraw_Blt( surface_type_e dsttype, LPRECT dstrect,
                    surface_type_e srctype, LPRECT srcrect,
                    DWORD flags, LPDDBLTFX fx )
{
    int result;

    LPDIRECTDRAWSURFACE4 lpDDS4_dst, lpDDS4_src;
    
    if( dsttype == primary_surface )
    {
	if( DirectDrawState.isoverlay )
	    lpDDS4_dst = DirectDrawState.overlay.surface_dx6;
	else
	    lpDDS4_dst = DirectDrawState.primary.surface_dx6;
    }
    else if( dsttype == secondary_surface )
    {
        lpDDS4_dst = DirectDrawState.secondary.surface_dx6;
    }
    else
    {
	lpDDS4_dst = DirectDrawState.overlay.surface_dx6;
    }

    if( srctype == primary_surface )
    {
        lpDDS4_src = DirectDrawState.primary.surface_dx6;
    }
    else if( srctype == secondary_surface )
    {
        lpDDS4_src = DirectDrawState.secondary.surface_dx6;
    }
    else if( srctype == overlay_surface )
    {
	lpDDS4_src = DirectDrawState.overlay.surface_dx6;
    }
    else
    {
	lpDDS4_src = NULL; /* For using BltStub to do rect-fills */
    }
    result = DirectDraw_BltStub( lpDDS4_dst, dstrect, lpDDS4_src, srcrect, flags, fx );
    return result;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_WaitForVerticalBlank( DWORD flags )
{
    HRESULT result;
    result = IDirectDraw4_WaitForVerticalBlank( DirectDrawState.directdraw.dd_dx6, flags, NULL );
    return result;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_EnumDisplayModes( DWORD flags, LPDDENUMMODESCALLBACK2 callback )
{
    HRESULT result;
    result = IDirectDraw4_EnumDisplayModes( DirectDrawState.directdraw.dd_dx6, flags, NULL, NULL, callback );    
    return result;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_FlipToGDISurface( void )
{
    HRESULT result = DDERR_GENERIC;
    if( DirectDrawState.initialized )
    {
	result = IDirectDraw4_FlipToGDISurface( DirectDrawState.directdraw.dd_dx6 );
    }
    return result;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_GetDC( HDC *hdc, surface_type_e surface )
{
    HRESULT result;
    if( surface == primary_surface )
        result = IDirectDrawSurface4_GetDC( DirectDrawState.primary.surface_dx6, hdc );
    else
        result = IDirectDrawSurface4_GetDC( DirectDrawState.secondary.surface_dx6, hdc );
    return result;
}

/*
 * FUNCTION:   
 *
 * PURPOSE:    
 *
 * PARAMETERS:
 *
 * RETURNS:
 *
 * NOTES:      
 *
 * HISTORY:
 *   1999.08.02  Brian King             Creation
 *
 */
HRESULT DirectDraw_ReleaseDC( HDC hdc, surface_type_e surface )
{
    HRESULT result;
    if( surface == primary_surface )
        result = IDirectDrawSurface4_ReleaseDC( DirectDrawState.primary.surface_dx6, hdc );
    else
        result = IDirectDrawSurface4_ReleaseDC( DirectDrawState.secondary.surface_dx6, hdc );
    return result;
}

HRESULT DirectDraw_UpdateOverlay( RECT sr, RECT dr )
{
    if (DirectDrawState.isoverlay && DirectDrawState.overlay.surface_dx6)
    {
	if ((drivercaps.dwCaps & DDCAPS_ALIGNBOUNDARYSRC) && drivercaps.dwAlignBoundarySrc)
	    sr.left = (sr.left + drivercaps.dwAlignBoundarySrc / 2) & ~(drivercaps.dwAlignBoundarySrc - 1);
	if ((drivercaps.dwCaps & DDCAPS_ALIGNSIZESRC) && drivercaps.dwAlignSizeSrc)
	    sr.right = sr.left + (sr.right - sr.left + drivercaps.dwAlignSizeSrc / 2) & ~(drivercaps.dwAlignSizeSrc - 1);
	if ((drivercaps.dwCaps & DDCAPS_ALIGNBOUNDARYDEST) && drivercaps.dwAlignBoundaryDest)
	    dr.left = (dr.left + drivercaps.dwAlignBoundaryDest / 2) & ~(drivercaps.dwAlignBoundaryDest - 1);    
	if ((drivercaps.dwCaps & DDCAPS_ALIGNSIZEDEST) && drivercaps.dwAlignSizeDest)
	    dr.right = dr.left + (dr.right - dr.left) & ~(drivercaps.dwAlignSizeDest - 1);
	return IDirectDrawSurface_UpdateOverlay( DirectDrawState.overlay.surface_dx6, &sr, DirectDrawState.primary.surface_dx6, &dr, overlayflags, &overlayfx);
    }
    return DD_OK;
}
